

import json
import os

from sqlalchemy import insert
from brave.api.models.core import t_namespace
import uuid

from brave.api.service.pipeline import get_pipeline_dir 

def save_namespace(conn,saveNamespace):
    # str_uuid = str(uuid.uuid4())     
    # saveNamespace["namespace_id"] = str_uuid
    stmt = t_namespace.insert().values(saveNamespace)
    conn.execute(stmt)

def find_namespace(conn,namespace_id):
    stmt = t_namespace.select().where(t_namespace.c.namespace_id == namespace_id)
    return conn.execute(stmt).mappings().first()

def update_namespace(conn,namespace_id,updateNamespace):
    stmt = t_namespace.update().where(t_namespace.c.namespace_id == namespace_id).values(updateNamespace)
    conn.execute(stmt)

def delete_namespace(conn,namespace_id):
    stmt = t_namespace.delete().where(t_namespace.c.namespace_id == namespace_id)

    conn.execute(stmt)

def list_namespace(conn): 
    stmt = t_namespace.select()
    return conn.execute(stmt).mappings().all()




# def import_namespace(conn,namespace_id,force=False):
#     pipeline_dir = get_pipeline_dir()
#     pipeline_dir = f"{pipeline_dir}/{namespace_id}"
#     with open(f"{pipeline_dir}/namespace.json","r") as f:
#         namespace_json = json.load(f)
#     find_namespace_db = find_namespace(conn,namespace_id)
#     if find_namespace_db:
#         if force:
#             update_stmt = t_namespace.update().where(t_namespace.c.namespace_id == namespace_id).values(namespace_json)
#             conn.execute(update_stmt)
#     else:
#         conn.execute(insert(t_namespace).values(namespace_json))
#     return namespace_json

# def write_namespace(namespace_id,saveNamespace):
#     pipeline_dir = get_pipeline_dir()
#     namespace = f"{pipeline_dir}/{namespace_id}"
#     if not os.path.exists(namespace):
#         os.makedirs(namespace)
#     with open(f"{namespace}/namespace.json","w") as f:
#         f.write(json.dumps(saveNamespace))

async def init_db(conn):
    namespace = find_namespace(conn,"default")
    if not namespace:
        namespace_dict = {"namespace_id":"default","name":"default","is_use":True}
        save_namespace(conn,namespace_dict)
        # write_namespace("default",namespace_dict)

    
def get_used_namespace(conn):
    stmt = t_namespace.select().where(t_namespace.c.is_use == True)
    return conn.execute(stmt).mappings().first()

def set_used_namespace(conn,namespace_id):
    stmt = t_namespace.update().where(t_namespace.c.is_use == True).values({"is_use":False})
    conn.execute(stmt)
    stmt = t_namespace.update().where(t_namespace.c.namespace_id == namespace_id).values({"is_use":True})
    conn.execute(stmt)
    return find_namespace(conn,namespace_id)


from typing import Optional
from pydantic import BaseModel

class SavePipeline(BaseModel):
    component_id: Optional[str]=None
    parent_component_id:Optional[str]=None
    # pipeline_id: Optional[str]=None
    relation_type: Optional[str]=None
    component_type: Optional[str]
    content: Optional[str]=None
    namespace:Optional[str]=None
    component_name:Optional[str]=None
    description:Optional[str]=None
    tags:Optional[str]=None
    category:Optional[str]=None
    img:Optional[str]=None
    container_id:Optional[str]=None
    order_index:Optional[int]=None
    sub_container_id:Optional[str]=None
    tags:Optional[str]=None

class SavePipelineRelation(BaseModel):
    relation_id:Optional[str]=None
    component_id: Optional[str]=None
    # pipeline_id: Optional[str]=None
    parent_component_id:Optional[str]=None
    # pipeline_key: Optional[str]=None
    relation_type: Optional[str]=None


class QueryPipeline(BaseModel):
    component_id: Optional[str]=None
    # parent_pipeline_id:Optional[str]=None
    # pipeline_key: Optional[str]=None
    order_index: Optional[int]=None
    component_type: Optional[str]=None
    content: Optional[str]=None
    namespace:Optional[str]=None


class PagePipelineQuery(BaseModel):
    page_number: Optional[int]=1
    page_size: Optional[int]=10
    component_type: Optional[str]=None
    keywords: Optional[str]=None
    # component_id: Optional[str]=None


class Pipeline(BaseModel):
    id:Optional[int]
    component_id: Optional[str]
    # parent_pipeline_id:Optional[str]
    # pipeline_key: Optional[str]
    order_index: Optional[int]
    component_type: Optional[str]
    component_name: Optional[str]
    content: Optional[str]
    namespace:Optional[str]

class QueryModule(BaseModel):
    module_type: Optional[str]
    module_name: Optional[str]=None
    component_id:Optional[str]
    file_type:Optional[str]=None
    # module_dir: Optional[str]=None


class SaveOrder(BaseModel):
    relation_id:Optional[str]=None
    order_index:Optional[int]=None

# class PipelineComponentsEdges(BaseModel):
#     source: str
#     sourceHandle: str
#     target: str
#     targetHandle: str



class SavePipelineComponentsEdges(BaseModel):
    component_id: str
    edges: str
    position:str

class InstallComponent(BaseModel):
    path: str





def parse_data(analysis_dict,database_dict,extra_dict,groups_name,groups,settings,metadata_form):
    sample_list = [{
                        "sample_name":item['sample_name'],
                        "sample_source":item['sample_source'],
                        "analysis_result_id":item['analysis_result_id'],
                        "sample_id":item['sample_id']
                    }
                   for v in analysis_dict.values() 
                   for item in v]
    
    return {
        "sample_list":sample_list,
        **extra_dict,
        **analysis_dict,
        **database_dict,
        "groups_name":groups_name,
        "groups":groups,
        "pipeline_dir":str(settings.PIPELINE_DIR),
        "metadata_form":metadata_form
    
    }


from sqlalchemy import and_

def build_where_clause(model_instance, table,field_list ):
    conditions = []
    for field, value in model_instance.dict(exclude_none=True).items():
        if field in field_list:
            column = getattr(table.c, field)
            conditions.append(column == value)
    return and_(*conditions)



from collections import defaultdict
import json
import os
import glob

from dependency_injector.wiring import Provide
from sqlalchemy.orm import aliased
from brave.api.config.config import get_settings
from pathlib import Path
import shutil
from fastapi import Depends, HTTPException
from brave.api.schemas.pipeline import PagePipelineQuery, SavePipeline,Pipeline,QueryPipeline,QueryModule,SaveOrder,SavePipelineComponentsEdges
from brave.api.config.db import get_engine
from brave.api.models.core import  t_pipeline_components, t_pipeline_components_edges, t_pipeline_components_relation
import importlib.resources as resources
from sqlalchemy import delete, select, and_, join, func,insert,update,or_
from datetime import datetime
from brave.api.enum.component_script import ScriptName
from sqlalchemy import  Column, Integer, String, Text, select, cast, null,text,case
from .notebook import generate_notebook
import brave.api.service.container_service as container_service

def get_pipeline_dir():
    settings = get_settings()
    return settings.PIPELINE_DIR

def get_pipeline_list():
    pipeline_dir =  get_pipeline_dir()
    pipeline_files = glob.glob(f"{pipeline_dir}/*/main.json")
    return pipeline_files
def get_module_name(item):
    pipeline_dir =  get_pipeline_dir()
    item_module = item.replace(f"{pipeline_dir}/","").replace("/",".")
    item_module = Path(item_module).stem 
    return item_module
    # return {os.path.basename(item).replace(".py",""):item_module} 
    # f'reads-alignment-based-abundance-analysis.py_plot.{module_name}'

def find_component_module(component,script_name:ScriptName) -> dict:
    # content = json.loads(component['content'])
    if "script_type" not in component:
        raise HTTPException(status_code=500, detail=f"script_type not found!")
    if script_name == ScriptName.main:
        script_type = component['script_type']
        if script_type == "nextflow":
            file_type = "nf"
        elif script_type == "python":
            file_type = "py"
        elif script_type == "shell":
            file_type = "sh"
        elif script_type == "r":
            file_type = "R"
        elif script_type == "jupyter":
            file_type = "ipynb"
        else:
            raise HTTPException(status_code=500, detail=f"script_type {script_type} not found!")
    else:
        file_type = "py"
    module_info = find_module(component['component_id'],script_name,file_type)
    if module_info:
        return module_info
    else:
        if script_name == ScriptName.main:
            create_file(component['component_id'],component['component_type'],file_type)
            module_info = find_module(component['component_id'],script_name,file_type)
            if not module_info:
                raise HTTPException(status_code=500, detail=f"组件{component.component_id}的{script_name.value}模块没有找到!")
            return module_info
        else:
            default_module = get_default_module(script_name)
            return default_module
        # else:
        #     raise HTTPException(status_code=500, detail=f"{script_name.value}没有找到默认模块!")





def find_module(module_dir,script_name:ScriptName,file_type):
    # if script_name == ScriptName.input_parse:
    # if not module_name:
    #     if  module_type == "script":
    #         module_name = "main"
    #     else:
    #         return get_default_module(module_type)

        # raise HTTPException(status_code=500, detail=f"模块名称不能为空!")
    # if module_name =="default":
       
    path_dict = get_all_module(file_type)
    # if module_name is None: 
    #     module_name = module_type
    if module_dir in path_dict:
        module_dict = path_dict[module_dir]
        if script_name.value in module_dict:
            module_info = module_dict[script_name.value]
            return module_info
    return None
        # return get_default_module(module_type)
        # raise HTTPException(status_code=500, detail=f"目录{module_type}: {module_dir}没有找到!")
    

    # if script_name.value not in py_module_dir:
    #     raise HTTPException(status_code=500, detail=f"目录{module_type}: {module_dir}/{script_name.value}没有找到!")

    
    # return py_module

def get_default_module(script_name:ScriptName):
    if script_name == ScriptName.input_parse:
        py_parse_analysis = resources.files("brave.parse").joinpath("py_parse_analysis.py")
        return {
            "module":"brave.parse.py_parse_analysis",
            "path":str(py_parse_analysis)
        }
    if script_name == ScriptName.output_parse:
        py_parse_analysis_result = resources.files("brave.parse").joinpath("py_parse_analysis_result.py")
        return {
            "module":"brave.parse.py_parse_analysis_result",
            "path":str(py_parse_analysis_result)
        }
    raise HTTPException(status_code=500, detail=f"{script_name.value}没有找到默认模块!")
def get_all_module(file_type):
    suffix = file_type
    # if module_type.startswith("py_"):
    #     suffix = "py"
    # else:
    #     suffix = "nf"
    pipeline_dir =  get_pipeline_dir()
    nextflow_list = glob.glob(f"{pipeline_dir}/*/*/*.{suffix}")
    result = {}
    for item in nextflow_list:
        
        parts = item.split(os.sep)
        dir_name = parts[-2]
        filename = os.path.basename(item).replace(f".{suffix}","")
        if dir_name not in result:
                result[dir_name] = {}
        
        item_module = get_module_name(item)
        result[dir_name][filename] = {
            "module":item_module,
            "path":item
        }
    
  
    # nextflow_dict = {os.path.basename(item).replace(".py",""):get_module_name(item) for item in nextflow_list}
    return result

def delete_wrap_pipeline_dir(pipeline_key):
    pipeline_dir =  get_pipeline_dir()
    src  =f"{pipeline_dir}/{pipeline_key}"
    dst  =f"{pipeline_dir}/bin"
    if not os.path.exists(dst):
        os.makedirs(dst)
    if os.path.exists(src):
        shutil.move(src,dst)

# def create_wrap_pipeline_dir(pipeline_key):
#     pipeline_dir = get_pipeline_dir()
#     img = f"{pipeline_dir}/{pipeline_key}/img"
#     nextflow = f"{pipeline_dir}/{pipeline_key}/nextflow"
#     py_parse_analysis = f"{pipeline_dir}/{pipeline_key}/py_parse_analysis"
#     py_parse_analysis_result = f"{pipeline_dir}/{pipeline_key}/py_parse_analysis_result"
#     py_plot = f"{pipeline_dir}/{pipeline_key}/py_plot"
    # dir_list = [img, nextflow,py_parse_analysis,py_parse_analysis_result,py_plot]
    # for item in dir_list:
    #     if not os.path.exists(item):
    #         os.makedirs(item) 

def create_file(component_id,component_type,file_type):
    pipeline_dir = get_pipeline_dir()
    pipeline_dir = f"{pipeline_dir}"

    content_text = ""
    if file_type == "py":
        with resources.files("brave.templete").joinpath(f"py_plot.py").open("r") as f:
            content_text = f.read()
    elif file_type == "nf":
        with resources.files("brave.templete").joinpath("nextflow.nf").open("r") as f:
            content_text = f.read()
    elif file_type == "R":
        with resources.files("brave.templete").joinpath("py_plot.R").open("r") as f:
            content_text = f.read()
    # if component_type == "pipeline":
    analysis_file = f"{pipeline_dir}/{component_type}/{component_id}/main.{file_type}"
    if not os.path.exists(analysis_file):
        dir_ = os.path.dirname(analysis_file)
        if not os.path.exists(dir_):
            os.makedirs(dir_) 
        if file_type=="ipynb":
            generate_notebook(analysis_file)
        else:
            with open(analysis_file,"w") as f:
                f.write(content_text)
            ipynb_path = os.path.dirname(analysis_file)
            generate_notebook(f"{ipynb_path}/main.ipynb")
    
    return analysis_file
        # if not os.path.exists(parseAnalysisModule):
        #     with open(parseAnalysisModule,"w") as f:
        #         f.write("")

    # if component_type == "software":
    #     # parseAnalysisModule = f"{pipeline_dir}/software/{component_id}/main.py"
    #     analysisPipline = f"{pipeline_dir}/software/{component_id}/main.{file_type}"
    #     # dir_list = [parseAnalysisModule,analysisPipline]
    #     # for item in dir_list:
    #     dir_ = os.path.dirname(analysisPipline)
    #     if not os.path.exists(dir_):
    #         os.makedirs(dir_) 
        
    #     if not os.path.exists(analysisPipline):
  
    #         with open(analysisPipline,"w") as f:
    #             f.write(content_text)
    #     return analysisPipline
    #     # if not os.path.exists(parseAnalysisModule):
    #     #     with resources.files("brave.templete").joinpath("py_parse_analysis.py").open("r") as f:
    #     #         content_text = f.read()
    #     #     with open(parseAnalysisModule,"w") as f:
    #     #         f.write(content_text)
        
    #     # if "parseAnalysisResultModule" in  content:
    #     #     parseAnalysisResultModule = content['parseAnalysisResultModule']
    #     #     for item in parseAnalysisResultModule:
    #     #         item_file = f"{pipeline_dir}/software/{component_id}/{item['module']}.py"
    #     #         dir_ = os.path.dirname(item_file)
    #     #         if not os.path.exists(dir_):
    #     #             os.makedirs(dir_) 
    #     #         if not os.path.exists(item_file):
    #     #             with resources.files("brave.templete").joinpath("py_parse_analysis_result.py").open("r") as f:
    #     #                 content_text = f.read()
    #     #             with open(item_file,"w") as f:
    #     #                 f.write(content_text)

    # if component_type == "script":
   

    #     py_plot = f"{pipeline_dir}/script/{component_id}/main.{file_type}"
    #     py_plot_dir = os.path.dirname(py_plot)
    #     if not os.path.exists(py_plot):
    #         if not os.path.exists(py_plot_dir):
    #             os.makedirs(py_plot_dir)
  
    #         with open(py_plot,"w") as f:
    #             f.write(content_text)

    #     nb_file = f"{pipeline_dir}/script/{component_id}/main.ipynb"
    #     if not os.path.exists(nb_file):
    #         generate_notebook(nb_file)

    #     return py_plot
    # raise HTTPException(status_code=500, detail=f"component_type {component_type} not create file!")


def find_component_by_id(conn,component_id):
    stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id ==component_id)
    find_pipeine = conn.execute(stmt).mappings().first()
    return find_pipeine

def find_pipeline_by_id(conn,component_id):
    stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id ==component_id)
    find_component = conn.execute(stmt).mappings().first()
    
    # component_type = find_component["component_type"]
    if find_component:
        find_component = dict(find_component)
        if find_component["container_id"]:
            find_container = container_service.find_container_by_id(conn,find_component["container_id"])
            find_component['container'] = find_container
        if find_component["sub_container_id"]:
            find_container = container_service.find_container_by_id(conn,find_component["sub_container_id"])
            find_component['sub_container'] = find_container

    # if find_component and find_component["server_container_id"]:
    #     find_component = dict(find_component)
    #     find_container = container_service.find_container_by_id(conn,find_component["container_id"])
    #     find_component['container'] = find_container
        # find_component['container_name'] = find_container["name"]
    return find_component

def find_component_by_parent_id(conn,parent_id,relation_type=None):
    stmt = (
        select(
            t_pipeline_components_relation,  # 关系表所有字段
            t_pipeline_components  # 组件表所有字段
        )
        .select_from(
            t_pipeline_components_relation.outerjoin(
                t_pipeline_components,
                t_pipeline_components_relation.c.component_id == t_pipeline_components.c.component_id
            )
        )
        .where(t_pipeline_components_relation.c.parent_component_id == parent_id)
    )

    if relation_type is not None:
        stmt = stmt.where(t_pipeline_components_relation.c.relation_type == relation_type)

    result = conn.execute(stmt).mappings().all()
    return result



def list_pipeline(conn,queryPipeline:QueryPipeline):
    stmt = t_pipeline_components.select() 
  
    conditions = []
    if queryPipeline.component_type is not None:
        conditions.append(t_pipeline_components.c.component_type == queryPipeline.component_type)

    # if analysisResultQuery.ids is not None:
    #     conditions.append(analysis_result.c.id.in_(analysisResultQuery.ids))
    # if analysisResultQuery.analysis_method is not None:
    #     conditions.append(analysis_result.c.analysis_method.in_(analysisResultQuery.analysis_method))
    # if analysisResultQuery.analysis_type is not None:
    #     conditions.append(analysis_result.c.analysis_type == analysisResultQuery.analysis_type)
    stmt= stmt.where(and_( *conditions))

    # stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_type ==queryPipeline.component_type)
    find_pipeine = conn.execute(stmt).fetchall()
    return find_pipeine

def format_pipeline_componnet_one(item):
    try:
        content = json.loads(item['content'])
        item = {**content,**{k:v for k,v in item.items() if k != 'content'}}
        if 'img' in item:
            if not item['img']:
                item['img'] = f"/brave-api/img/pipeline.jpg"
    except Exception as e:
        print("component json error",json.dumps(item,indent=4))
 
        # else:
    return item

def page_pipeline(conn,query:PagePipelineQuery):
    stmt =select(
        t_pipeline_components,
    ) 
    
   

    conditions = []
    if query.component_type is not None:
        conditions.append(t_pipeline_components.c.component_type == query.component_type)

    if query.keywords:
        keyword_pattern = f"%{query.keywords}%"
        conditions.append(
            or_(
                t_pipeline_components.c.component_name.ilike(keyword_pattern),
                t_pipeline_components.c.tags.ilike(keyword_pattern),
                t_pipeline_components.c.description.ilike(keyword_pattern)
            )
        )

    stmt = stmt.where(and_(*conditions))
    count_stmt = select(func.count()).select_from(t_pipeline_components).where(and_(*conditions))

    stmt = stmt.offset((query.page_number - 1) * query.page_size).limit(query.page_size)
    find_pipeline = conn.execute(stmt).mappings().all()
    find_pipeline = [dict(item) for item in find_pipeline]
    find_pipeline = [format_pipeline_componnet_one(item) for item in find_pipeline]

    total = conn.execute(count_stmt).scalar()
    return {
        "items": find_pipeline,
        "total":total,
        "page_number":query.page_number,
        "page_size":query.page_size
    }



 
def datetime_converter(o):
    if isinstance(o, datetime):
        return o.isoformat()
    raise TypeError(f"Type {type(o)} not serializable")



def find_by_relation_id(conn,relation_id:str):
    stmt = t_pipeline_components_relation.select().where(t_pipeline_components_relation.c.relation_id == relation_id)
    find_pipeline = conn.execute(stmt).mappings().first()
    return find_pipeline



def find_component_by_container_id(conn,container_id):
    stmt = t_pipeline_components.select().where(t_pipeline_components.c.container_id == container_id)
    return conn.execute(stmt).mappings().all()

def get_child_depend_component(conn, component_id):
    
    stmt = select(
        t_pipeline_components_relation.c.relation_id,
        t_pipeline_components_relation.c.relation_type,
        t_pipeline_components.c.component_id,
        t_pipeline_components.c.component_type,
        t_pipeline_components.c.component_name,
        t_pipeline_components.c.description,
        ).select_from(
            t_pipeline_components_relation.outerjoin(
                t_pipeline_components,
                t_pipeline_components_relation.c.component_id == t_pipeline_components.c.component_id,
            )
        ).where(
            t_pipeline_components_relation.c.parent_component_id == component_id,
        )
    


    return conn.execute(stmt).mappings().all()

def get_parent_depend_component(conn, component_id):

    stmt = select(
        t_pipeline_components_relation.c.relation_id,
        t_pipeline_components_relation.c.relation_type,
        t_pipeline_components.c.component_id,   
        t_pipeline_components.c.component_type,
        t_pipeline_components.c.component_name,
        t_pipeline_components.c.description,
    ).select_from(
        t_pipeline_components_relation.outerjoin(
            t_pipeline_components,
            t_pipeline_components_relation.c.parent_component_id == t_pipeline_components.c.component_id,
        )
    ).where(
        t_pipeline_components_relation.c.component_id == component_id,
    )
    return conn.execute(stmt).mappings().all()




def get_child_component_count(conn, component_id,relation_type):
    stmt = select(func.count()).select_from(t_pipeline_components_relation).where(
        t_pipeline_components_relation.c.parent_component_id == component_id,
        t_pipeline_components_relation.c.relation_type == relation_type,
    )
    return conn.execute(stmt).scalar()


def save_order(conn,saveOrderList:list[SaveOrder]):
    for saveOrder in saveOrderList:
        stmt = update(t_pipeline_components_relation).where(t_pipeline_components_relation.c.relation_id == saveOrder.relation_id).values(order_index=saveOrder.order_index)
        conn.execute(stmt)



def save_pipeline_components_edges(conn,savePipelineComponentsEdges:SavePipelineComponentsEdges):
    # stmt = delete(t_pipeline_components_edges).where(t_pipeline_components_edges.c.pipeline_id == savePipelineComponentsEdges.component_id)
    # conn.execute(stmt)
    # for pipeline_components_edge in savePipelineComponentsEdges.pipeline_components_edges:
    #     stmt = insert(t_pipeline_components_edges).values(
    #         source=pipeline_components_edge.source,
    #         sourceHandle=pipeline_components_edge.sourceHandle,
    #         target=pipeline_components_edge.target,
    #         targetHandle=pipeline_components_edge.targetHandle,
    #         pipeline_id=savePipelineComponentsEdges.component_id
    #     )
    #     conn.execute(stmt)
        
    update_stmt = update(t_pipeline_components).where(
        t_pipeline_components.c.component_id == savePipelineComponentsEdges.component_id
        ).values(position=savePipelineComponentsEdges.position,
        edges=savePipelineComponentsEdges.edges  )
    conn.execute(update_stmt)





def get_pipeline_v2(conn,name,component_type="pipeline"):
    base = (select(
        t_pipeline_components.c.component_id,
        t_pipeline_components.c.component_type,
        t_pipeline_components.c.install_key,
        t_pipeline_components.c.content,
        t_pipeline_components.c.component_name,
        t_pipeline_components.c.position,
        t_pipeline_components.c.edges,
        t_pipeline_components.c.tags,
        t_pipeline_components.c.description,
        cast(null(), String(255)).label("relation_type"),
        cast(null(), String(255)).label("parent_component_id"),
        cast(null(), String(255)).label("order_index"),
        cast(null(), String(255)).label("relation_id"),
    ).where(
        t_pipeline_components.c.component_id == name,
        t_pipeline_components.c.component_type == component_type,
    ).cte(name="base", recursive=True))


    # 递归 CTE 定义
    # cte = base.cte(name="full_pipeline", recursive=True)

    tp1 = aliased(t_pipeline_components)
    rel = t_pipeline_components_relation
    # fp = aliased(cte)  # 引用递归CTE自身
    base_alias = base.alias()

    recursive = select(
        tp1.c.component_id,
        tp1.c.component_type,
        tp1.c.install_key,
        tp1.c.content,
        tp1.c.component_name,
        tp1.c.position,
        tp1.c.tags,
        tp1.c.edges,
        tp1.c.description,
        rel.c.relation_type,
        rel.c.parent_component_id,
        rel.c.order_index,
        rel.c.relation_id,
    ).select_from(
        tp1.join(rel, tp1.c.component_id == rel.c.component_id)
        .join(base_alias, rel.c.parent_component_id == base_alias.c.component_id)
    )

    # # 合并 base 和 recursive，生成完整的递归CTE
    cte = base.union_all(recursive) #.cte(name="full_pipeline", recursive=True)

    # # 最终查询排序
    final_query = select(cte).order_by(
        case((cte.c.order_index == None, 1), else_=0),
        cte.c.order_index
        # func.coalesce(cte.c.order_index, cte.c.relation_id),
    )
    # 执行查询
    # with get_engine().begin() as conn:
    # [dict(item) for item in data]
    data = conn.execute(final_query).mappings().all()
    if  len(data) < 0:
        raise HTTPException(status_code=500, detail=f"{name}没有找到!")  
    id_to_node = {(item["component_id"], item["relation_type"]): get_one_data(item) for item in data}
    children_map = defaultdict(list)
    for item in data:
        parent_id = item.get("parent_component_id")
        if parent_id:
            # parent_key = (parent_id, item["relation_id"])  # parent 的唯一标识
            child_key = (item["component_id"], item["relation_type"])
            children_map[parent_id].append(child_key)
    # 获取根 pipeline_id
    root_item  = next((item for item in data if item["component_type"] == component_type),None)
    if not root_item:
        raise HTTPException(status_code=500, detail=f"{name}没有找到!")  
    # root_key = (root_item["component_id"], None)
    if component_type == "software":
        result = build_software_structure(id_to_node,children_map,root_item)
    elif component_type == "pipeline":
        result = build_pipeline_structure(id_to_node,children_map,root_item)
    elif component_type == "file":
        result = build_file_structure(id_to_node,children_map,root_item)
    else:
        raise HTTPException(status_code=500, detail=f"{component_type}没有结构解析!")  
    return result

def get_one_data(item):
    try:
        content = json.loads(item["content"])
    except Exception as e:
        print("component json error",json.dumps(dict(item),indent=4))
        print("component json error",e)
        content = {}
    # end try
    
    item = {k:v for k,v in item.items() if k!="content"}
    return { **content,**item }

def build_file_structure(id_to_node,children_map,root_item):
    
    item = {**root_item, "downstreamAnalysis": []}
    for ds_id in children_map.get(item["component_id"], []):
        downstream = id_to_node[ds_id]
        item["downstreamAnalysis"].append(downstream)
    return item

def  build_software_structure(id_to_node,children_map,root_item):
    # result = {**root_item}
    item = {**root_item}
    input_files = []
    output_files = []
    for sub_id in children_map.get(root_item["component_id"], []):
        sub = id_to_node[sub_id]
        sub_content = sub
        if sub["relation_type"] == "software_input_file":
            input_files.append(sub_content)
        elif sub["relation_type"] == "software_output_file":
            sub_out = {**sub_content, "downstreamAnalysis": []}
            for ds_id in children_map.get(sub["component_id"], []):
                downstream = id_to_node[ds_id]
                sub_out["downstreamAnalysis"].append(downstream)
            output_files.append(sub_out)    
    return {**item, "inputFile":input_files, "outputFile":output_files}

def build_pipeline_structure(id_to_node,children_map,root_item):
    # node = id_to_node[pid]
    result = {**root_item}
    items = []
    for child_id in children_map.get(root_item["component_id"], []):
        child = id_to_node[child_id]
        content = child
        if child["component_type"] == "software":
            item = {**content}
            input_files = []
            output_files = []
            for sub_id in children_map.get(child_id[0], []):
                sub = id_to_node[sub_id]
                sub_content = sub
                if sub["relation_type"] == "software_input_file":
                    input_files.append(sub_content)
                elif sub["relation_type"] == "software_output_file":
                    sub_out = {**sub_content, "downstreamAnalysis": []}
                    for ds_id in children_map.get(sub["component_id"], []):
                        downstream = id_to_node[ds_id]
                        sub_out["downstreamAnalysis"].append(downstream)
                    output_files.append(sub_out)


            if input_files:
                item["inputFile"] = input_files
            if output_files:
                item["outputFile"] = output_files
            if "upstreamFormJson" in content:
                item["upstreamFormJson"] = content["upstreamFormJson"]
            items.append(item)
    if items:
        # result["items"] = items
        # item["databases"] for item in items if "databases" in item
        result["databases"] = [
             db for item in items if "databases" in item for db in item["databases"]
        ]
        result["upstreamFormJson"] = [
            db for item in items if "upstreamFormJson" in item and isinstance(item["upstreamFormJson"],list) and len(item["upstreamFormJson"]) > 0   for db in item["upstreamFormJson"]
        ]
        result["software"] = items
        if items and len(items) > 0 and "inputFile" in items[0]:
            result["inputFile"] = items[0]["inputFile"]
        else:
            result["inputFile"] = []
    return result


  
def update_component_description(conn, component_id,description):
    stmt = t_pipeline_components.update().where(t_pipeline_components.c.component_id ==component_id).values(description=description)
    conn.execute(stmt)
    # pass







def get_components(conn, component_id):
    base = (
        select(t_pipeline_components)
        .where(t_pipeline_components.c.component_id == component_id)
        .cte(name="base", recursive=True)
    )
    tp1 = aliased(t_pipeline_components)
    rel = t_pipeline_components_relation
    base_alias = base.alias()

    # Use INNER JOINs for the recursive part to comply with MySQL's CTE restrictions
    recursive = (
        select(tp1)
        .select_from(
            tp1.join(rel, tp1.c.component_id == rel.c.component_id)
               .join(base_alias, rel.c.parent_component_id == base_alias.c.component_id)
        )
    )

    cte = base.union_all(recursive)
    final_query = select(cte)
    data = conn.execute(final_query).mappings().all()
    return data

def get_component_relation(conn,component_id):
    base = (select(
        t_pipeline_components_relation
    ).where(
        t_pipeline_components_relation.c.parent_component_id == component_id,
    ).cte(name="base", recursive=True))
    tr1 = aliased(t_pipeline_components_relation)

    base_alias = base.alias()

    recursive = select(tr1).select_from(
        tr1.join(base_alias, tr1.c.parent_component_id == base_alias.c.component_id)
    )

    cte = base.union_all(recursive) 
    final_query = select(cte)
    data = conn.execute(final_query).mappings().all()
    return data

def delete_component_file(component):
    component_type = component["component_type"]
    component_id = component["component_id"]
    pipeline_dir = get_pipeline_dir()
    pipeline_dir = f"{pipeline_dir}/{component_type}/{component_id}"
    if os.path.exists(pipeline_dir):
        shutil.rmtree(pipeline_dir)

def write_component_json(component_id):
    with get_engine().begin() as conn:
        component_list = get_components(conn,component_id)
        component_relation_list =  get_component_relation(conn,component_id)
        container_id_list = [ item['container_id'] for item in component_list if item['container_id'] is not None ]
        container_list = container_service.find_by_container_ids(conn,container_id_list)
    
    current_component = [item for item in component_list if item.component_id == component_id]
    if not current_component:
        raise HTTPException(status_code=500, detail=f"组件{component_id}没有找到!")
    current_component = current_component[0]
    component_type = current_component["component_type"]
    pipeline_dir = get_pipeline_dir()
    pipeline_dir = f"{pipeline_dir}/{component_type}/{component_id}"
    if not os.path.exists(pipeline_dir):
        os.makedirs(pipeline_dir)
    
    pipeline_component_file = f"{pipeline_dir}/pipeline_component.json"
    pipeline_component_relation_file = f"{pipeline_dir}/pipeline_component_relation.json"
    with open(pipeline_component_file,"w") as f:
        json.dump([ {k:v for k,v in item.items() if k!="id"} for item in component_list],f, default=datetime_converter)
    with open(pipeline_component_relation_file,"w") as f:
        json.dump([ {k:v for k,v in item.items() if k!="id" and k!="created_at" and k!="updated_at"} for item in component_relation_list],f, default=datetime_converter)    

    container_file = f"{pipeline_dir}/container.json"
    container_list = [ {k:v for k,v in item.items() if k!="id" and k!="created_at" and k!="updated_at"} for item in container_list]
    with open(container_file,"w") as f:
        json.dump(container_list,f)
    
    install_file = f"{pipeline_dir}/install.json"
    with open(install_file,"w") as f:
        json.dump({
            "component_type":component_type,
            "component_id":component_id,
            "component_name":current_component["component_name"],
            "img":os.path.basename(current_component["img"]) if current_component["img"] else "",
        },f)
    

        # component = t_pipeline_components.select().where(t_pipeline_components.c.component_id ==component_id)
    print(f"export pipeline_component_file: {pipeline_component_file}!")
    print(f"export pipeline_component_relation_file: {pipeline_component_relation_file}!")
    print(f"export container_file: {container_file}!")

    pass



def import_component_relation(conn,path,force=False):
    # pipeline_dir = get_pipeline_dir()
    with open(f"{path}/pipeline_component_relation.json","r") as f:
        find_pipeline = json.load(f)
    for item in find_pipeline:
        find_pipeline_component_relation = find_by_relation_id(conn,item['relation_id'])
        if find_pipeline_component_relation:
            if force:
                update_stmt = update(t_pipeline_components_relation).where(t_pipeline_components_relation.c.relation_id == item['relation_id']).values(item)
                conn.execute(update_stmt)
        else:
            conn.execute(insert(t_pipeline_components_relation).values(item))   
def import_component(conn,path,force=False):

    with open(f"{path}/pipeline_component.json","r") as f:
        find_pipeline = json.load(f)
    for item in find_pipeline:
        find_pipeline_component = find_pipeline_by_id(conn,item['component_id'])
        if find_pipeline_component:
            if force:
                update_stmt = update(t_pipeline_components).where(t_pipeline_components.c.component_id == item['component_id']).values(item)
                conn.execute(update_stmt)
        else:
            conn.execute(insert(t_pipeline_components).values(item))   


# Find components by a list of component IDs
def find_by_component_ids(conn,component_ids):
    stmt = t_pipeline_components.select().where(t_pipeline_components.c.component_id.in_(component_ids))
    find_pipeline = conn.execute(stmt).mappings().all()
    return find_pipeline



from brave.api.config.db import get_engine
from fastapi import HTTPException
from brave.api.models.core import analysis_result, samples,analysis,t_pipeline_components,t_project
from brave.api.schemas.analysis_result import AnalysisResult,AnalysisResultQuery
from sqlalchemy import and_, desc, select,case
import json
import uuid
from collections import defaultdict

def get_analysis_result_metadata(item):
    if item["metadata"]:
        metadata = json.loads(item["metadata"])
        prefix = ""
        if item["sample_source"]:
            prefix = f"{item['sample_source']}-"
        metadata = {k:f"{prefix}{v}" for k,v in metadata.items() if v is not None}
        item = {**metadata,**item}
        del item["metadata"]
    return item

def find_analyais_result_groupd_by_component_ids(conn,component_ids,projectList):
    result_dict =  find_analyais_result(conn,AnalysisResultQuery(component_ids=component_ids,projectList=projectList))
    result_dict = [get_analysis_result_metadata(item) for item in result_dict]
            
        # if item["metadata_form"]:
        #     item["metadata_form"] = json.loads(item["metadata_form"])
    grouped = defaultdict(list)
    for item in result_dict:
        item["label"] = item["sample_name"]
        item["value"] = item["id"]
        grouped[item["component_id"]].append(item)
    return grouped

def find_analyais_result(conn,analysisResultQuery:AnalysisResultQuery):
    stmt = analysis_result.select() 
    if analysisResultQuery.querySample:
        stmt =  select(
            analysis_result, 
            samples.c.sample_name,
            # samples.c.sample_group,
            # samples.c.sample_id,
            samples.c.metadata,
            analysis.c.analysis_name,
            t_pipeline_components.c.component_name.label("component_name"),
            # t_pipeline_components.c.label.label("component_label"),
            # t_pipeline_components.c.name.label("analysis_method"),
            t_project.c.project_name.label("project_name")
            # t_project.c.project_id.label("project_id")
            # t_project.c.metadata_form.label("metadata_form")

            ) 
        stmt = stmt.select_from(
            analysis_result.outerjoin(samples,samples.c.sample_id==analysis_result.c.sample_id)
            .outerjoin(analysis,analysis.c.analysis_id==analysis_result.c.analysis_id)
            .outerjoin(t_pipeline_components,t_pipeline_components.c.component_id==analysis_result.c.component_id)
            .outerjoin(t_project,t_project.c.project_id==analysis_result.c.project)
            )
        # stmt = stmt.where(samples.c.project == analysisResultQuery.project)
    # if analysisResultQuery.queryAnalysis:
    #     if not analysisResultQuery.querySample:
    #         stmt = select(analysis_result)
    #     stmt = stmt.add_columns(
    #         analysis.c.analysis_name
    #     )
    #     stmt = stmt.get_final_froms()[0].select_from(analysis_result.outerjoin(analysis,analysis.c.analysis_id==analysis_result.c.analysis_id))
    
    conditions = []
    if analysisResultQuery.project is not None:
        conditions.append(analysis_result.c.project == analysisResultQuery.project)
    if analysisResultQuery.ids is not None:
        conditions.append(analysis_result.c.id.in_(analysisResultQuery.ids))
    if analysisResultQuery.analysis_method is not None:
        conditions.append(analysis_result.c.analysis_method.in_(analysisResultQuery.analysis_method))
    if analysisResultQuery.analysis_type is not None:
        conditions.append(analysis_result.c.analysis_type == analysisResultQuery.analysis_type)
    if analysisResultQuery.component_ids is not None:
        conditions.append(analysis_result.c.component_id.in_(analysisResultQuery.component_ids))
    if analysisResultQuery.component_id is not None:
        conditions.append(analysis_result.c.component_id == analysisResultQuery.component_id)
    if analysisResultQuery.projectList is not None:
        conditions.append(analysis_result.c.project.in_(analysisResultQuery.projectList))


    stmt= stmt.where(and_( *conditions))
    
    if analysisResultQuery.ids :
        case_order = case(
            {id_: index for index, id_ in enumerate(analysisResultQuery.ids)},
            value=analysis_result.c.id,
            else_=len(analysisResultQuery.ids)
        )
        stmt = stmt.order_by(case_order)
    
    if analysisResultQuery.component_ids:
        case_order = case(
            {id_: index for index, id_ in enumerate(analysisResultQuery.component_ids)},
            value=analysis_result.c.component_id,
            else_=len(analysisResultQuery.component_ids)
        )
        stmt = stmt.order_by(case_order)
    
    

    result  = conn.execute(stmt)
    # result = result.fetchall()
    rows = result.mappings().all()
    result_dict = [dict(item) for item in rows ]
    # result_dict = [AnalysisResult(**row) for row in rows]
    # result_dict = []
    for item in result_dict:
        if item['content_type']=="json" and not isinstance(item['content'], dict):
            item['content'] = json.loads(item['content'])
            
        #     result.append(row)
        # # rows = result.mappings().all()
        # pass
    return result_dict

def model_dump_one(item):
    if item.get("content_type")=="json" and  isinstance(item.get("content"), dict):
        if item["metadata"]:
            metadata = json.loads(item["metadata"])
            item = {**metadata,**item}
            del item["metadata"]

        return{
            **{k:v for k,v in item.items() if k!="content"},
            **item['content']
            # 'content':item['content']
        }
    return item

def find_analyais_result_by_ids( conn,value):
    ids = value
    if not isinstance(value,list):
        ids = [value]
    analysis_result = find_analyais_result(conn,AnalysisResultQuery(ids=ids))
    analysis_result = [model_dump_one(item) for item in analysis_result]
    if len(analysis_result)!=len(ids):
        raise HTTPException(status_code=500, detail="数据存在问题!")
    if not isinstance(value,list) and len(analysis_result)==1:
        return analysis_result[0]
    else:
        return analysis_result

def find_analysis_result_exist(conn,component_id,file_name,project):
    stmt = analysis_result.select().where(and_(
        analysis_result.c.component_id == component_id,
        analysis_result.c.file_name == file_name,
        analysis_result.c.project == project
    ))
    result = conn.execute(stmt).mappings().first()
    return result 


            
def add_analysis_result(conn,analysis_result_dict):
    print(f"添加分析结果: {analysis_result_dict['file_name']}")
    analysis_result_dict['analysis_result_id'] = str(uuid.uuid4())
    stmt = analysis_result.insert().values(analysis_result_dict)
    conn.execute(stmt)
def update_analysis_result(conn,analysis_result_id,analysis_result_dict):
    print(f"更新分析结果: {analysis_result_dict['file_name']}")
    stmt = analysis_result.update().where(analysis_result.c.id==analysis_result_id).values(analysis_result_dict)
    conn.execute(stmt)  

def find_by_analysis_result_id(conn,analysis_result_id):
    stmt = analysis_result.select().where(analysis_result.c.analysis_result_id==analysis_result_id)
    result = conn.execute(stmt).mappings().first()
    return result   

def update_sample_id(conn,analysis_result_id,sample_id):
    stmt = analysis_result.update().where(analysis_result.c.analysis_result_id==analysis_result_id).values({"sample_id":sample_id})
    conn.execute(stmt)
        
    
def find_analysis_result_by_analysis_id(conn,analysis_id):
    stmt = analysis_result.select().where(analysis_result.c.analysis_id==analysis_id)
    result = conn.execute(stmt).mappings().all()
    return result

def find_analysis_result_in_analysis_id(conn,analysis_id_list):
    stmt = select(analysis_result)
    # stmt = stmt.select_from(analysis_result.outerjoin(samples,samples.c.sample_id==analysis_result.c.sample_id))
    stmt = stmt.where(analysis_result.c.analysis_id.in_(analysis_id_list))
    result = conn.execute(stmt).mappings().all()
    return result   
    # with get_db_session() as db:
    #     if len(res) >0:
    #         # print(res[0])
    #         if len(res[0]) == 4:
    #             for analysis_key,software,content_type,content in res:
    #                 update_or_save_result(analysis_key,analysis_key, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id)
    #         elif len(res[0]) == 5:
    #             for analysis_key,sample_name,software,content_type,content in res:
    #                 update_or_save_result(analysis_key,sample_name, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id)


# def update_or_save_result(analysis_key,sample_name, software, content_type, content, db, project, verison, analysis_method,analysis_name,analysis_id):
#         sampleAnalysisResult = db.query(SampleAnalysisResult) \
#         .filter(and_(SampleAnalysisResult.analysis_method == analysis_method,\
#                 SampleAnalysisResult.analysis_version == verison, \
#                 SampleAnalysisResult.analysis_key == analysis_key, \
#                 SampleAnalysisResult.project == project \
#             )).first()
#         if sampleAnalysisResult:
#             sampleAnalysisResult.sample_name = sample_name
#             sampleAnalysisResult.content = content
#             sampleAnalysisResult.sample_key=sample_name
#             sampleAnalysisResult.content_type = content_type
#             sampleAnalysisResult.analysis_name = analysis_name
#             sampleAnalysisResult.analysis_id = analysis_id
#             sampleAnalysisResult.analysis_type="upstream"
#             # sampleAnalysisResult.log_path = log_path
#             sampleAnalysisResult.software = software
#             db.commit()
#             db.refresh(sampleAnalysisResult)
#             print(">>>>更新: ",sample_name, software, content_type)
#         else:
#             sampleAnalysisResult = SampleAnalysisResult(analysis_method=analysis_method, \
#                 analysis_version=verison, \
#                 sample_name=sample_name, \
#                 content_type=content_type, \
#                 analysis_name=analysis_name, \
#                 analysis_key=analysis_key, \
#                 analysis_id=analysis_id, \
#                 analysis_type="upstream", \
#                 # log_path=log_path, \
#                 software=software, \
#                 project=project, \
#                 sample_key=sample_name, \
#                 content=content \
#                     )
#             db.add(sampleAnalysisResult)
#             db.commit()
#             print(">>>>新增: ",sample_name, software, content_type)




import asyncio
import inspect
from importlib import import_module
from importlib.resources import files
from functools import lru_cache
class ListenerFilesService:
    def __init__(self):
        self.listener_files = self._load_listener_files()

    def _load_listener_files(self):
        """
        加载监听器文件列表
        """
        listener_files = files("brave.api.listener")
        return [
            item.stem
            for item in listener_files.iterdir()
            if item.is_file() and item.name.endswith(".py") and item.name != "__init__.py" and item.name.startswith("file")
        ]

    async def execute_listener(self, func, args):
        """
        执行监听器的回调函数
        """
        if isinstance(self.listener_files, list) and len(self.listener_files) > 0:
            for name in self.listener_files:
                full_module = f"brave.api.listener.{name}"
                mod = import_module(full_module)
                if hasattr(mod, func):
                    run_func = getattr(mod, func)
                    if inspect.iscoroutinefunction(run_func):
                        asyncio.create_task(run_func(**args))
                    else:
                        await asyncio.to_thread(run_func, **args)

    def get_listener_files(self):
        return self.listener_files
    
    def add_listener_file(self,listener_file):
        self.listener_files.append(listener_file)
    
    def remove_listener_file(self,listener_file):
        self.listener_files.remove(listener_file)

@lru_cache()
def get_listener_files_service():
    return ListenerFilesService()


from fastapi import Request, APIRouter
from brave.api.core.event import IngressEvent
from brave.api.core.routers.ingress_event_router import IngressEventRouter

class HTTPIngress:
    def __init__(self, router:IngressEventRouter):
        self.router = router

    def create_endpoint(self):
        async def endpoint(request: Request):
            msg = await request.json()
        
            evnet_str = msg.get("ingress_event")
            event = IngressEvent(evnet_str)
            if not event:
                event = IngressEvent.NEXTFLOW_EXECUTOR_EVENT
                print(f"[HTTP] Unknown event type '{event}'", msg)
                return
            data = msg.get("data")
            await self.router.dispatch(event,data)
            return {"status": "ok"}
        return endpoint
        # app.include_router(router)



from abc import ABC, abstractmethod
import asyncio
from collections import defaultdict
from ctypes import Union
from typing import Any, Callable, Awaitable, Coroutine, Optional, Type, TypeVar, Generic, Union, Dict, Set

from pydantic import BaseModel

from brave.api.core.event import WorkflowEvent
E = TypeVar("E")  # 事件类型（如 Enum）
F = TypeVar("F", bound=Callable[..., Any])  
# Callback = Union[Callable[[dict], None], Callable[[dict], Coroutine],Callable[[str,dict], Coroutine]]
P = TypeVar("P", bound=BaseModel)

class BaseEventRouter(ABC, Generic[E,F]):
    def __init__(self):
        # self._handlers: Dict[str, Callable[[dict], Awaitable]] = {}
        self._handlers: dict[E, set[F]] = defaultdict(set)
        # self._payload_types: Dict[E, Type[BaseModel]] = {}

    def on_event2(self, event: E):
        def decorator(func: F):
            self._handlers[event].add(func)
            return func
        return decorator


    def on_event(self, event_type: E):
        def decorator(func:F):
            self._handlers[event_type].add(func)
            # if payload_type:
            #     self._payload_types[event_type] = payload_type
            # return func
        return decorator

    def register_handler(self, event: E, handler: F):
        self._handlers[event].add(handler)

    async def dispatch(self, event: E, *payload: BaseModel):
        handlers = self._handlers.get(event, [])
        # expected_type = self._payload_types.get(event)
        # if  expected_type:
        #     if not isinstance(payload, expected_type):
        #         raise TypeError(f"[EventRouter] Expected payload type {expected_type}, got {type(payload)}")

        for handler in handlers:
            if asyncio.iscoroutinefunction(handler):
                await handler(*payload)
            else:
                handler(*payload)
        else:
            print(f"[BaseEventRouter] No handler for event '{event}'")
            
    # async def dispatch(self,event:E, msg: dict):
    #     # event = msg.get("workflow_event")
    #     # if not event:
    #     #     print("[EventRouter] No 'workflow_event' in message", msg)
    #     #     return
    #     handlers = self._handlers.get(event)
    #     if handlers:
    #         for handler in handlers:
    #             if asyncio.iscoroutinefunction(handler):
    #                 await handler(msg)
    #             else:
    #                 handler(msg)
    #     else:
    #         print(f"[EventRouter] No handler for event '{event}'")



import json
import os
import asyncio
import psutil
import signal
import subprocess
import threading
from pathlib import Path
from typing import Dict

from brave.api.core.evenet_bus import EventBus
from brave.api.executor.base import JobExecutor
from brave.api.executor.models import JobSpec
from brave.api.config.config import Settings, get_settings
from brave.api.core.routers_name import RoutersName
from brave.api.core.event import AnalysisExecutorEvent
from brave.api.schemas.analysis import AnalysisId
# PROCESS_FILE = Path("/tmp/local_processes.json")  # 可配置

class LocalExecutor(JobExecutor):
    def __init__(self,event_bus:EventBus):
        self.event_bus = event_bus
        self.settings:Settings = get_settings()
        self.PROCESS_FILE = self._process_file()
        self.loop = asyncio.get_event_loop()

        self._processes: Dict[str, int] = {}
        self._load_processes()
   
    def _process_file(self):
        return f"{self.settings.BASE_DIR}/local_processes.json"

    def _save_processes(self):
        try:
            with open(self.PROCESS_FILE, "w") as f:
                json.dump(self._processes, f)
        except Exception as e:
            print(f"[LocalExecutor] Failed to persist process map: {e}")

    def _load_processes(self):
        if Path(self.PROCESS_FILE).exists():
            try:
                with open(self.PROCESS_FILE, "r") as f:
                    data = json.load(f)
                    for job_id, pid in data.items():
                        try:
                            proc = psutil.Process(pid)
                            if proc.is_running():
                                self._processes[job_id] = pid
                                # 重启后的 wait 监听
                                threading.Thread(
                                    target=self._wait_and_cleanup_factory(pid, job_id),
                                    daemon=True
                                ).start()
                        except Exception:
                            continue  # 跳过无效进程
            except Exception as e:
                print(f"[LocalExecutor] Failed to load persisted processes: {e}")

    def _wait_and_cleanup_factory(self, pid: int, job_id: str):
        def _wait():
            try:
                psutil.Process(pid).wait()
            except Exception:
                pass
            self._processes.pop(job_id, None)
            self._save_processes()
            analysis_id = AnalysisId(analysis_id=job_id)
            asyncio.run_coroutine_threadsafe(self.event_bus.dispatch(RoutersName.ANALYSIS_EXECUTER_ROUTER,AnalysisExecutorEvent.ON_ANALYSIS_COMPLETE,analysis_id),self.loop)
            print(f"[JobManager] Process {job_id} with PID {pid} exited and cleaned.")
        return _wait

    async def _do_submit_job(self, job_spec) -> str:
        if job_spec.job_id in self._processes:
            pid = self._processes[job_spec.job_id]
            try:
                proc = psutil.Process(pid)
                if proc.is_running():
                    raise Exception(f"Analysis already running with PID={pid}")
            except psutil.NoSuchProcess:
                pass  # 继续重新提交

        proc = subprocess.Popen(
            job_spec.command,
            cwd=job_spec.output_dir,
            start_new_session=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

        self._processes[job_spec.job_id] = proc.pid
        self._save_processes()

        threading.Thread(
            target=self._wait_and_cleanup_factory(proc.pid, job_spec.job_id),
            daemon=True
        ).start()

        return str(proc.pid)

    def stop_job(self, job_id: str) -> None:
        pid = self._processes.get(job_id)
        if pid:
            try:
                proc = psutil.Process(pid)
                if proc.is_running():
                    os.killpg(pid, signal.SIGTERM)
                    print(f"[JobManager] Terminated process group of PID {pid}")
                else:
                    raise Exception(f"Process {pid} not running")
            finally:
                self._processes.pop(job_id, None)
                self._save_processes()

    def get_logs(self, job_id: str) -> str:  
        return super().get_logs(job_id)


# brave/api/ingress/manager.py
from fastapi import FastAPI
from .factory import  create_ingress, IngressMode
from .interfaces.base_ingress import BaseMessageIngress
from brave.api.core.routers.ingress_event_router import IngressEventRouter
from .http_ingress import HTTPIngress

class IngressManager:
    def __init__(
        self, 
        ingress_event_router: IngressEventRouter ,
        event_mode: IngressMode = IngressMode.STREAM, 
        uds_path: str = "/tmp/brave.sock"):
    
        # self.ingress: BaseMessageIngress = create_ingress(event_mode, uds_path, ingress_event_router)
        self.ingress = create_ingress(event_mode, uds_path, ingress_event_router)

    async def start(self):
        if isinstance(self.ingress, BaseMessageIngress):
            await self.ingress.start()
    
    def create_endpoint(self):
        if isinstance(self.ingress, HTTPIngress):
            return self.ingress.create_endpoint()
        return None

    # def register_http(self, app):
    #     if isinstance(self.ingress, HTTPIngress):
    #         self.ingress.register(app)



from enum import Enum


class RoutersName(str, Enum):
    INGRESS_EVENT_ROUTER = "ingress_event_router"
    WORKFLOW_EVENT_ROUTER = "workflow_event_router"
    WATCHFILE_EVENT_ROUTER = "watchfile_event_router"
    ANALYSIS_EVENT_ROUTER = "analysis_event_router"
    ANALYSIS_EXECUTER_ROUTER = "analysis_executer_router"
    FILE_WATCHER_EVENT_ROUTER = "file_watcher_event_router"
    SSE_EVENT_ROUTER = "sse_event_router"
    HTTP_EVENT_ROUTER = "http_event_router"
    UDS_EVENT_ROUTER = "uds_event_router"
    UDS_STREAM_EVENT_ROUTER = "uds_stream_event_router"
    ANALYSIS_RESULT_ROUTER="analysis_result_router"